@@ -30,7 +30,9 @@
 
 #include <variant>
 
-namespace mega {
+namespace mega
+{
+using namespace std::literals;
 
 // helper class for categorizing transfers for upload/download queues
 struct TransferCategory
@@ -152,6 +154,8 @@ struct MEGA_API Transfer : public FileFingerprint
     MegaClient* client;
     int tag;
 
+    void setProgresscompleted(const m_off_t p, const bool append = false);
+
     // signal failure.  Either the transfer's slot or the transfer itself (including slot) will be deleted.
     void failed(const Error&, TransferDbCommitter&, dstime = 0);
 
@@ -177,6 +181,8 @@ struct MEGA_API Transfer : public FileFingerprint
     // temp URLs for upload/download data.  They can be cached.  For uploads, a new url means any previously uploaded data is abandoned.
     // downloads can have 6 for raid, 1 for non-raid.  Uploads always have 1
     std::vector<string> tempurls;
+    uint8_t discardedTempUrlsSize{};
+    static constexpr m_time_t TEMPURL_TIMEOUT_TS{172500};
 
     // context of the async fopen operation
     unique_ptr<AsyncIOContext> asyncopencontext;
@@ -210,12 +216,16 @@ struct MEGA_API Transfer : public FileFingerprint
     // whether it is a Transfer for support (i.e., an upload for the Support team)
     bool isForSupport() const;
 
-    // whether the transfer is a Sync upload transfer
-    bool mIsSyncUpload = false;
-
     // Add stats for this transfer to the MEGAclient. The client must be valid at this point.
     bool addTransferStats();
 
+    void collectAndPrintTransferStatsIfLimitReached();
+
+    void discardTempUrlsIfNoDataDownloadedOrTimeoutReached(const direction_t transferDirection,
+                                                           const m_time_t currentTime);
+
+    void adjustNonRaidedProgressIfNowIsRaided();
+
 private:
     FileDistributor::TargetNameExistsResolution toTargetNameExistsResolution(CollisionResolution resolution);
 };
@@ -281,82 +291,74 @@ private:
 
 /**
  * @brief Represents unused connection in a Raided Streaming Transfer.
- * This struct stores the connection number, the reason why is unused, and the number of backoff
- * retries (in case of reason is UN_TEMP_ERR) before it can be reused again
+ * This struct stores the connection number and the reason why is unused
+ *
+ * @note A bandwidth overquota error (509) cannot affect only a specific raided part, it applies to
+ * the entire transfer, so it will be considered as an invalid error in this struct.
+ *
  */
 struct UnusedConn
 {
 public:
-    static constexpr unsigned int defaultBackoffRetries{5};
-    static constexpr std::array httpErrCodes = {
-        0, // NETWORK ERROR
-        403, // FORBIDDEN
-        404, // NOT FOUND
-        429, // TOO MANY REQUESTS
-        500, // INTERNAL SERVER ERROR
-        503, // SERVICE UNAVALIABLE
+    /**
+     * @enum unusedReason
+     * @brief Represents the reason why unused connection has been set as unused.
+     * - INVALID REASON: Connection failed due 509, which should be managed by retrying entire
+     * transfer with a backoff (DirectReadSlot::retryOnError should not be called with this error)
+     * - UN_NOT_ERR: Unused connection has not failed yet, so it can be switched by another
+     * connection if it's needed
+     * - UN_DEFINITIVE_ERR: Unused connection has failed with a definitive error, so it cannot be
+     * reused anymore.
+     * - UN_TEMPORARY_ERR: Unused connection has failed by a temporary error (i.e HttpStatus 200 and
+     * CURLcode: 56), so it can be switched by another connection if it's needed
+     */
+    enum UnusedReason
+    {
+        UN_INVALID = 0,
+        UN_NOT_ERR = 1,
+        UN_DEFINITIVE_ERR = 2,
+        UN_TEMPORARY_ERR = 3,
     };
 
-    enum unusedReason
+    /**
+     * @enum connReplacementReason
+     * @brief Represents the reason why a connection has been replaced by unused one.
+     * - CONN_SPEED_SLOWEST_PART replaced part is the slowest one in comparison with the rest of
+     * parts
+     *
+     * - TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD replaced part is the slowest one and transfer mean
+     * speed is below minstreamingrate or replaced part speed is below min speed threshold
+     *
+     * - ON_RAIDED_ERROR replaced part has failed with an HTTP error.
+     */
+    enum ConnReplacementReason
     {
-        UN_INVALID = 0, // INVALID REASON
-        UN_NOT_ERR = 1, // CONNECTION CAN BE USED
-        UN_TEMP_ERR = 2, // CONNECTION CAN BE USED AFTER A BACKOFF
-        UN_DEFINITIVE_ERR = 3, // CONNECTION CANNOT BE USED ANYMORE
-        UN_RETRY_IMMEDIATELY = 4, // ENTIRE TRANSFER MUST BE RETRIED IMMEDIATELY
+        CONN_SPEED_SLOWEST_PART = 0,
+        TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD = 1,
+        ON_RAIDED_ERROR = 2,
+        ON_RECOVERABLE_RAIDED_ERROR = 3,
     };
 
     /**
      * @brief Returns an unusedReason given a HTTP status code.
      *
-     * @note Currently this method never returns UN_DEFINITIVE_ERR UN_DEFINITIVE_ERR is only set
-     * if same connection has failed more than once and failedParts has exceeded
-     * MAX_FAILED_RAIDED_PARTS
-     *
      * @return An unusedReason given a HTTP status code.
      */
-    static unusedReason getReasonFromHttpStatus(const int httpstatus)
+    static UnusedReason getReasonFromHttpStatus(const int httpstatus, const unsigned errCode)
     {
+        constexpr unsigned CURLE_OK_CODE = 0;
         switch (httpstatus)
         {
-            case 404:
-            case 429:
-                return UN_RETRY_IMMEDIATELY;
-            case 0:
-            case 403:
-            case 503:
-            case 500:
-                return UN_TEMP_ERR;
-            default:
+            case 200:
+                return errCode == CURLE_OK_CODE ? UN_NOT_ERR : UN_TEMPORARY_ERR;
+            case 509:
                 assert(false);
                 return UN_INVALID;
+            default:
+                return UN_DEFINITIVE_ERR;
         }
     }
 
-    /**
-     * @brief Checks if the provided HTTP status code is handled.
-     *
-     * @note if this method returns false entire transfer should be retried
-     * as there's no specific code to handle that HTTP status
-     *
-     * @return true if the HTTP status code is valid, otherwise returns false.
-     */
-    static bool isHandledHttpStatus(const int httpstatus)
-    {
-        return std::find(httpErrCodes.begin(), httpErrCodes.end(), httpstatus) !=
-               httpErrCodes.end();
-    }
-
-    /**
-     * @brief Checks if mReason is an error reason
-     *
-     * @return true if the reason is `UN_TEMP_ERR` or `UN_DEFINITIVE_ERR`, otherwise returns false.
-     */
-    static bool isErrReason(const unusedReason& reason)
-    {
-        return reason == UN_TEMP_ERR || reason == UN_DEFINITIVE_ERR;
-    }
-
     /**
      * @brief Gets the number of the unused connection.
      *
@@ -365,27 +367,15 @@ public:
     size_t getNum() const;
 
     /**
-     * @brief Checks if the current reason is a temporary error.
-     *
-     * This method checks if the unused connection is a temporary error reason (`UN_TEMP_ERR`).
-     *
-     * @return true if the reason is `UN_TEMP_ERR`, otherwise returns false.
+     * @brief Checks if unused connection can be reused (mReason is not an error reason)
+     * @return true if unused connection can be reused, otherwise returns false.
      */
-    bool isTempErrReason() const;
-
-    /**
-     * @brief Checks if mReason is an error reason
-     * @see isErrReason(const unusedReason& reason)
-     * @return true if the reason is an error reason, otherwise returns false.
-     */
-    bool isErrReason() const;
+    bool CanBeReused() const;
 
     /**
      * @brief Sets the unused connection info.
      *
-     * This method sets the unused state of the connection, the connection number and clears
-     * mNumBackoffRetries. If reason is UN_TEMP_ERR, it also reset mNumBackoffRetries to
-     * defaultBackoffRetries
+     * This method sets the unused state of the connection
      *
      * @param num The number of connection
      * @param reason The reason for marking the connection as unused
@@ -393,41 +383,24 @@ public:
      * @return true if the reason is valid and the connection state was updated successfully,
      *         false if the reason is invalid.
      */
-    bool setUnused(const size_t num, const unusedReason reason);
+    bool setUnused(const size_t num, const UnusedReason reason);
 
     /**
      * @brief Resets the unused connection state.
      */
     void clear();
-
-    /**
-     * @brief Decrements the temporary backoff retries counter.
-     *
-     * This method decrements the number of backoff retries, only if mReason is UN_TEMP_ERR and if
-     * it's greater than 0, otherwise it returns. If mNumBackoffRetries reaches 0 after decrement,
-     * it also resets mReason to UN_NOT_ERR
-     *
-     * If mNumBackoffRetries is greater that 0 and mReason is UN_TEMP_ERR it means that this
-     * connection cannot be used temporarily, until mNumBackoffRetries is 0 and mReason is
-     * UN_NOT_ERR
-     *
-     * @see searchAndDisconnectSlowestConnection
-     */
-    void decTempBackoffRetries();
-
-private:
     /**
      * @brief Checks if reason provided by param is a valid unusedReason.
      *
      * @return true if the reason is valid, otherwise returns false.
      */
-    static bool isValidUnusedReason(unusedReason reason)
+    static bool isValidUnusedReason(const UnusedReason reason)
     {
-        return reason == UN_NOT_ERR || reason == UN_TEMP_ERR || reason == UN_DEFINITIVE_ERR;
+        return reason == UN_NOT_ERR || reason == UN_DEFINITIVE_ERR || reason == UN_TEMPORARY_ERR;
     }
 
-    unsigned int mNumBackoffRetries{};
-    unusedReason mReason{UN_NOT_ERR};
+private:
+    UnusedReason mReason{UN_NOT_ERR};
     size_t mNum{};
 };
 
@@ -451,6 +424,11 @@ public:
      *      Constants     *
     \* ===================*/
 
+    /**
+     * @brief Default unused connection index
+     */
+    static constexpr unsigned DEFAULT_UNUSED_CONN_INDEX = 0;
+
     /**
     *   @brief Time interval to recalculate speed and mean speed values.
     *
@@ -507,11 +485,28 @@ public:
     static constexpr unsigned DEFAULT_MIN_COMPARABLE_THROUGHPUT = MAX_DELIVERY_CHUNK;
 
     /**
-    *   @brief Max times a DirectReadSlot is allowed to detect a slower connection and switch it to the unused one.
-    *
-    *   @see DirectReadSlot::searchAndDisconnectSlowestConnection()
-    */
-    static constexpr unsigned MAX_SLOW_CONNECTION_SWITCHES = 6;
+     *   @brief Max times a DirectReadSlot is allowed to switch the unused connection for another
+     * connection detected as slow with respect to the others.
+     *
+     *   @see DirectReadSlot::searchAndDisconnectSlowestConnection()
+     */
+    static constexpr unsigned MAX_CONN_SWITCHES_SLOWEST_PART = 6;
+
+    /**
+     *   @brief Max times a DirectReadSlot is allowed switch unused connection by another connection
+     * detected slower than min threshold
+     *
+     *   @see DirectReadSlot::detectAndManageSlowRaidedConns() and
+     * DirectReadSlot::onLowSpeedRaidedTransfer()
+     */
+    static constexpr unsigned MAX_CONN_SWITCHES_BELOW_SPEED_THRESHOLD = 1;
+
+    /**
+     *   @brief Max times a DirectReadSlot is allowed switch unused connection by another connection
+     * when a recoverable error is received (i.e HttpStatus 200 and CURLcode: 56)
+     *
+     */
+    static constexpr unsigned MAX_CONN_SWITCHES_RECOVERABLE_ERR = 1;
 
     /**
     *   @brief Requests are sent in batch, and no connection is allowed to request the next chunk until the other connections have finished fetching their current one.
@@ -537,24 +532,30 @@ public:
     static constexpr m_off_t SLOWEST_TO_FASTEST_THROUGHPUT_RATIO[2] { 4, 5 };
 
     /**
-     * @brief Max times a failed raided part of a DirectRead, is allowed to be replaced by another
-     * one.
+     * @brief Max simultaneus slow raided parts of a DirectRead allowed
      *
-     * @see DirectReadSlot::retry
+     * @see DirectReadSlot::watchOverDirectReadPerformance
      */
-    static constexpr int MAX_FAILED_RAIDED_PARTS = 1;
+    static constexpr unsigned MAX_SIMULTANEOUS_SLOW_RAIDED_CONNS = 1;
 
     /**
-     * @brief Max different failed raided parts of a DirectRead allowed
+     * @brief Timeout to reset connection switches counters.
      *
-     * @see DirectReadSlot::retryOnError
+     * During a streaming transfer, we may perform RAIDED parts replacements due to different
+     * reasons (failed part, slow mean speed). These replacements can be done just a limited number
+     * of times. However, for long streaming transfers, we need to reset those counters to discard
+     * punctual connectivity issues.
+     *
+     * @see DirectReadSlot::doio
      */
-    static constexpr int MAX_DIFFERENT_FAILED_RAIDED_CONNS = 1;
+    static constexpr std::chrono::seconds CONNECTION_SWITCHES_LIMIT_RESET_TIME = 300s;
 
     /* ===================*\
      *      Methods       *
     \* ===================*/
 
+    bool isRaidedTransfer() const;
+
     /**
      * @brief Retries the entire direct read transfer upon a failure.
      *
@@ -564,18 +565,31 @@ public:
      * @param err The error code that caused the failure
      * @param timeleft The time after which the transfer is retried
      */
-    void retryEntireTransfer(const Error& e, dstime timeleft = 0);
+    void retryEntireTransfer(const Error& e, const dstime timeleft = 0);
 
     /**
-     * @brief Manages low speed raided direct read transfer
+     * @brief Identifies slow connections under minimum threshold and determines the slowest one.
      *
-     * The function finds the slowest connection, and replaces it by unused connection if posible,
-     * Otherwise entire transfer is retried.
+     * This function iterates through all active and completed connections, checks
+     * their throughput against a minimum threshold, and identifies those that are
+     * considered too slow. Additionally, it determines the slowest one among them.
      *
-     * @note This function is called from watchOverDirectReadPerformance that monitors the speed of
-     * the entire raided transfer.
+     * @return A pair containing:
+     *         - A set of indices representing connections that are too slow.
+     *         - The index of the slowest connection (or invalid index if no slow conns exist).
      */
-    void onLowSpeedRaidedTransfer();
+    std::pair<std::set<size_t>, size_t> searchSlowConnsUnderThreshold();
+
+    void resetConnSwitchesCounters(const std::chrono::steady_clock::time_point& now);
+
+    /**
+     * @brief Reset all connection switches counters if timeout
+     * (CONNECTION_SWITCHES_LIMIT_RESET_TIME) has expired
+     */
+    void resetConnSwitchesCountersIfTimeoutExpired();
+
+    // Returns true if any raided Req has failed, otherwise returns false
+    bool isAnyRaidedPartFailed() const;
 
     /**
     *   @brief Main i/o loop (process every HTTP req from req vector).
@@ -588,7 +602,9 @@ public:
      * @brief Manages a HTTP req failure filtering by httpstatus and performing required action (i.e
      * retry HTTP req)
      */
-    void onFailure(std::unique_ptr<HttpReq>& req, const size_t connectionNum);
+    void onFailure(std::unique_ptr<HttpReq>& req,
+                   const size_t connectionNum,
+                   const unsigned errCode);
 
     /**
     *   @brief Flag value getter to check if a given request is allowed to request a further chunk.
@@ -621,18 +637,28 @@ public:
     bool resetConnection(size_t connectionNum = 0);
 
     /**
-    *   @brief Calculate throughput for a given connection: relation of bytes per millisecond.
-    *
-    *   Throughput is updated every time a new chunk is submitted to the transfer buffer.
-    *   Throughput values are reset when a new request starts.
-    *
-    *   @param connectionNum Connection index in mReq vector.
-    *   @return Connection throughput: average number of bytes fetched per millisecond.
-    *
-    *   @see DirectReadSlot::detectSlowestStartConnection()
-    *   @see DirectReadSlot::calcThroughput()
-    *   @see DirectReadSlot::mThroughPut
-    */
+     * @brief Retrieves the minimum speed per connection in Bytes per second.
+     *
+     * This method calculates the minimum allowed speed in Bytes per second for a connection,
+     * taking into account whether it's a streaming RAID transfer, and the limits configured in the
+     * client (check minstreamingrate).
+     *
+     * @return The minimum speed per connection in Bytes per second.
+     */
+    unsigned getMinSpeedPerConnBytesPerSec() const;
+
+    /**
+     *   @brief Calculate throughput for a given connection: relation of bytes per millisecond.
+     *
+     *   Throughput is updated every time a new chunk is submitted to the transfer buffer.
+     *   Throughput values are reset when a new request starts.
+     *
+     *   @param connectionNum Connection index in mReq vector.
+     *   @return Connection throughput: average number of bytes fetched per millisecond.
+     *
+     *   @see DirectReadSlot::calcThroughput()
+     *   @see DirectReadSlot::mThroughPut
+     */
     m_off_t getThroughput(size_t connectionNum) const;
 
     /**
@@ -640,14 +666,11 @@ public:
      *
      * This method attempts to retry a DirectRead transfer. If the transfer is non RAIDED,
      * it directly triggers a retry. If it's RAIDED, it replaces that part with unused RAID
-     * connection, and retries only that part.
-     *
-     * @note In case of RAIDED transfer and mFailedRaidedParts has reached or exceeded
-     * MAX_FAILED_RAIDED_PARTS, entire transfer will be retried
+     * connection (if possible), and retries only that part.
      *
      * @param connectionNum The connection number to retry.
      */
-    void retryOnError(const size_t connectionNum, const int httpstatus);
+    void retryOnError(const size_t connectionNum, const int httpstatus, const unsigned errCode);
 
     /**
      * @brief Check if there are in-flight requests
@@ -657,26 +680,97 @@ public:
     bool exitDueReqsOnFlight() const;
 
     /**
-     * @brief Replace unused connection by connectionNum
-     * @param connectionNum The connection number
-     * @return a pair of booleans, where first represents if connection could be replaced at
-     * DirectReadBufferManager level, and second one if new "usable" connection could be reset fine
+     * @brief Determines if the unused connection can be reused.
+     *
+     * @return `true` if the unused connection can be reused, `false` otherwise.
      */
-    std::pair<bool, bool> replaceUnusedConnection(size_t connectionNum);
+    bool unusedConnectionCanBeReused();
 
     /**
-    *   @brief Search for the slowest connection and switch it with the actual unused connection.
-    *
-    *   This method is called between requests:
-    *   If WAIT_FOR_PARTS_IN_FLIGHT is true, this ensures to compare among all the connections before they POST again.
-    *   If WAIT_FOR_PARTS_IN_FLIGHT is false, any connection with a REQ_INFLIGHT status will be ignored for comparison purposes.
-    *
-    *   @param connectionNum Connection index in mReq vector.
-    *   @return True if the slowest connection has been detected and switched with the actual unused connection, False otherwise.
-    *   @see DirectReadSlot::MIN_COMPARABLE_THROUGHPUT
-    *   @see DirectReadSlot::MAX_SLOW_CONECCTION_SWITCHES
-    */
-    bool searchAndDisconnectSlowestConnection(size_t connectionNum = 0);
+     * @brief Replace connectionNum by unused connection when there are requests in flight.
+     * - This method decrements the number of requests in flight as necessary if the
+     * newUnusedConnection can be replaced by the currently unused one.
+     *
+     * @note: this method internally calls DirectReadSlot::replaceConnectionByUnused to perform
+     * connection replacement
+     *
+     * @param newUnusedConnection The connection number to be replaced by unused one
+     * @param reason Reason of replacement
+     * - UnusedConn::CONN_SPEED_SLOWEST_PART: replaced part is the slowest one
+     * - UnusedConn::TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD: replaced part is the slowest one AND
+     * transfer mean speed is below minstreamingrate OR replaced part speed is below min speed
+     * threshold
+     * - UnusedConn::ON_RAIDED_ERROR replaced part has failed due a Http err
+     *
+     * @param unusedReason reason to be set at new unused connection. See UnusedReason enum
+     *
+     */
+    void replaceConnectionByUnusedInflight(
+        const size_t newUnusedConnection,
+        const UnusedConn::ConnReplacementReason replamecementReason,
+        const UnusedConn::UnusedReason unusedReason);
+
+    /**
+     * @brief Replace connectionNum by unused connection
+     *
+     * @param newUnusedConnection The connection number to be replaced by unused one
+     * @param reason Reason of replacement
+     * - UnusedConn::CONN_SPEED_SLOWEST_PART: replaced part is the slowest one
+     * - UnusedConn::TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD: replaced part is the slowest one AND
+     * transfer mean speed is below minstreamingrate OR replaced part speed is below min speed
+     * threshold
+     * - UnusedConn::ON_RAIDED_ERROR replaced part has failed due a Http err
+     *
+     * @param unusedReason reason to be set at new unused connection. See UnusedReason enum
+     *
+     * @return True if connection has been replaced by unused, false otherwise
+     */
+    bool replaceConnectionByUnused(const size_t newUnusedConnection,
+                                   const UnusedConn::ConnReplacementReason replamecementReason,
+                                   const UnusedConn::UnusedReason unusedReason);
+
+    /**
+     * @brief Identifies the slowest and fastest connections (ignoring unused connection)
+     *
+     * @param connectionNum The index of the connection to compare others against.
+     * @return A `std::pair` where:
+     *         - The first element is the index of the slowest connection.
+     *         - The second element is the index of the fastest connection.
+     *         If no valid comparison can be made, both values will be set to `mReqs.size()`.
+     */
+    std::pair<size_t, size_t> searchSlowestAndFastestConns(const size_t connectionNum) const;
+
+    /**
+     * @brief Determines if the slowest connection can be replaced by the unused connection.
+     *
+     * @param connectionNum The index of the connection being evaluated.
+     * @param slowestConnection The index of the slowest connection identified.
+     * @param fastestConnection The index of the fastest connection identified.
+     * @return `true` if the slowest connection can be switched out for the unused connection,
+     *         `false` otherwise.
+     */
+    bool slowestConnTooSlowVsFastest(const size_t connectionNum,
+                                     const size_t slowestConnection,
+                                     const size_t fastestConnection) const;
+
+    /**
+     * @brief Search for the slowest connection and switch it with the actual unused connection.
+     *
+     * This method is intented to select fastest 5 connections (after all 5 raided parts finished a
+     * chunk)
+     *
+     * This method is called between requests:
+     * If WAIT_FOR_PARTS_IN_FLIGHT is true, this ensures to compare among all the connections
+     * before they POST again. If WAIT_FOR_PARTS_IN_FLIGHT is false, any connection with a
+     * REQ_INFLIGHT status will be ignored for comparison purposes.
+     *
+     * @param connectionNum Connection index in mReq vector.
+     * @return True if the slowest connection has been detected and switched with the actual
+     * unused connection, False otherwise.
+     * @see DirectReadSlot::DEFAULT_MIN_COMPARABLE_THROUGHPUT
+     * @see DirectReadSlot::MAX_CONN_SWITCHES_BELOW_SPEED_THRESHOLD
+     */
+    bool searchAndDisconnectSlowestConnection(const size_t connectionNum = 0);
 
     /**
      * @brief Checks if the minimum comparable throughput is met for a specific connection.
@@ -685,12 +779,18 @@ public:
      * @return true if the throughput for the specified connection meets the minimum comparable
      * threshold, otherwise returns false
      */
-    bool isMinComparableThroughputForThisConnection(const size_t connectionNum)
+    bool isMinComparableThroughputForThisConnection(const size_t connectionNum) const
     {
         return mThroughput[connectionNum].second &&
                mThroughput[connectionNum].first >= mMinComparableThroughput;
     }
 
+    /**
+     * @brief Check if all requests are inflight or in ready status
+     * @return True if all requests are inflight or in ready status
+     */
+    bool areAllReqsReadyOrInFlight();
+
     /**
     *   @brief Decrease counter for requests with REQ_INFLIGHT status
     *
@@ -713,6 +813,19 @@ public:
     */
     bool increaseReqsInflight();
 
+    /**
+     * @brief Returns a pair of [transfer min speed, transfer mean speed]
+     * @param dsSinceLastWatch Ds since watchOverDirectReadPerformance was executed last time
+     * @return a pair of [transfer min speed, transfer mean speed]
+     */
+    std::pair<int, m_off_t> getMinAndMeanSpeed(const dstime dsSinceLastWatch);
+
+    /**
+     * @brief Resets the watchdog associated variables that are used to perform some checkups based
+     * on elapsed time since last check and received data
+     */
+    void resetWatchdogPartialValues();
+
     /**
     *   @brief Calculate speed and mean speed for DirectRead aggregated operations.
     *
@@ -729,7 +842,7 @@ public:
      *
      * @return true if connection is done, otherwise returns false
      */
-    bool isConnectionDone(const size_t connectionNum);
+    bool isConnectionDone(const size_t connectionNum) const;
 
     /**
     *   @brief Builds a DirectReadSlot attached to a DirectRead object.
@@ -738,52 +851,6 @@ public:
     */
     DirectReadSlot(DirectRead*);
 
-    /**
-     * @brief Increments the count of failed raided parts.
-     *
-     * This method increments the counter of failed raided parts, and stores connNum at
-     * mFailedRaidedParts.
-     *
-     * @param connNum The number of connection
-     * @return The updated count of failed raided parts.
-     */
-    std::size_t recordFailedRaidedPart(size_t connNum)
-    {
-        mFailedRaidedParts.emplace(connNum);
-        return ++mFailedRaidedPartsCounter;
-    }
-
-    /**
-     * @brief decrements the count of failed raided parts.
-     *
-     * @return The updated count of failed raided parts.
-     */
-    std::size_t decrementFailedRaidedPartsCounter()
-    {
-        return --mFailedRaidedPartsCounter;
-    }
-
-    /**
-     * @brief gets the number of different failed raided parts (connections).
-     *
-     * @return The number of different failed raided parts.
-     */
-    std::size_t getDifferentRaidedPartsFailed() const
-    {
-        return mFailedRaidedParts.size();
-    }
-
-    /**
-     * @brief Resets the count of failed raided parts.
-     *
-     * This method sets the counter of failed raided parts to 0.
-     */
-    void clearFailedRaidedParts()
-    {
-        mFailedRaidedParts.clear();
-        mFailedRaidedPartsCounter = 0;
-    }
-
     /**
     *   @brief Destroy DirectReadSlot and stop any pendant operation.
     *
@@ -829,23 +896,13 @@ private:
     std::vector<std::unique_ptr<HttpReq>> mReqs;
 
     /**
-     * @brief Number of failed raided parts of a DirectRead.
-     * @see DirectReadSlot::retry
+     *   @brief Vector of pairs of <Bytes downloaded> and <Total milliseconds> for throughput
+     * calculations.
+     *
+     *   Values are reset by default between different chunk requests.
+     *
+     *   @see DirectReadSlot::DEFAULT_MIN_COMPARABLE_THROUGHPUT
      */
-    size_t mFailedRaidedPartsCounter{};
-
-    /**
-     * @brief Set of failed raided parts of a DirectRead.
-     */
-    std::set<size_t> mFailedRaidedParts;
-
-    /**
-    *   @brief Pair of <Bytes downloaded> and <Total milliseconds> for throughput calculations.
-    *
-    *   Values are reset by default between different chunk requests.
-    *
-    *   @see DirectReadSlot::MIN_COMPARABLE_THROUGHPUT
-    */
     std::vector<std::pair<m_off_t, m_off_t>> mThroughput;
 
     /**
@@ -862,19 +919,49 @@ private:
     std::chrono::steady_clock::time_point mSlotStartTime;
 
     /**
-    *   @brief Unused connection due to slowness.
-    *
-    *   This value is used for detecting the slowest start connection and further search and disconnect new slowest connections.
-    *   It must be synchronized with RaidBufferManager value, which is the one to be cached (so we keep it if reseting the DirectReadSlot).
-    */
+     * @brief Timeout to reset all connection switch counters.
+     *
+     * @see DirectReadSlot::resetConnSwitchesCountersIfTimeoutExpired
+     */
+    std::chrono::steady_clock::time_point mConnectionSwitchesLimitLastReset;
+
+    /**
+     *   @brief Unused connection due to slowness.
+     *
+     *   This value is used for detecting the slowest start connection and further search and
+     * disconnect new slowest connections. It must be synchronized with RaidBufferManager value,
+     * which is the one to be cached (so we keep it if reseting the DirectReadSlot).
+     */
     UnusedConn mUnusedConn;
 
     /**
-    *   @brief Current total switches done, i.e.: the slowest connection being switched with the unused connection.
-    *
-    *   @see DirectReadSlot::MAX_SLOW_CONNECTION_SWITCHES
-    */
-    unsigned mNumSlowConnectionsSwitches;
+     * @brief Current total of switches due to performance, i.e., the slowest part being switched
+     * with an unused connection (comparative logic among parts).
+     *
+     * @see DirectReadSlot::MAX_CONN_SWITCHES_SLOWEST_PART
+     */
+    unsigned mNumConnSwitchesSlowestPart;
+
+    /**
+     * @brief Current total of switches due to slow connections, i.e., a connection performing below
+     * the defined min speed threshold (minstrate).
+     *
+     * @see DirectReadSlot::MAX_CONN_SWITCHES_BELOW_SPEED_THRESHOLD
+     */
+    unsigned mNumConnSwitchesBelowSpeedThreshold;
+
+    /**
+     * @brief Current total of switches due to recoverable errors (i.e HttpStatus 200 and CURLcode:
+     * 56)
+     *
+     * @see DirectReadSlot::MAX_CONN_SWITCHES_RECOVERABLE_ERR
+     */
+    unsigned mNumConnSwitchesRecoverableError;
+
+    /**
+     * @brief maps connection id (raided part id) to number of slow speed detections
+     */
+    std::map<size_t, unsigned> mNumConnDetectedBelowSpeedThreshold;
 
     /**
     *   @brief Current flag value for waiting for the other connects to finish their TCP requests before any other connection is allowed to request the next chunk.
@@ -890,6 +977,13 @@ private:
     */
     unsigned mNumReqsInflight;
 
+    /**
+     * @brief Flag that indicates whether the DirectReadSlot::mNumReqsInflight counter has been
+incremented after processing the unused connection.
+@see DirectReadSlot::increaseReqsInFlight
+     */
+    bool mUnusedConnIncrementedInFlightReqs{false};
+
     /**
     *   @brief Speed controller instance.
     */
@@ -920,17 +1014,17 @@ private:
     unsigned mMaxChunkSize;
 
     /**
-    *   @brief Min submitted bytes needed for a connection to be throughput-comparable with others.
-    *
-    *   This value is set on global delivery throughput.
-    *   Ex:
-    *       1. Raid file, each connection submits 1MB.
-    *       2. Delivery chunk size from combined data is 5MB -> min comparable throughtput until next deliver will be 5MB.
-    *
-    *   @see detectSlowestStartConnection()
-    *   @see searchAndDisconnectSlowestConnection()
-    *   @see processAnyOutputPieces()
-    */
+     *   @brief Min submitted bytes needed for a connection to be throughput-comparable with others.
+     *
+     *   This value is set on global delivery throughput.
+     *   Ex:
+     *       1. Raid file, each connection submits 1MB.
+     *       2. Delivery chunk size from combined data is 5MB -> min comparable throughtput until
+     * next deliver will be 5MB.
+     *
+     *   @see searchAndDisconnectSlowestConnection()
+     *   @see processAnyOutputPieces()
+     */
     m_off_t mMinComparableThroughput;
 
     /**
@@ -948,6 +1042,69 @@ private:
      *   Private aux methods  *
     \* =======================*/
 
+    /**
+     * @brief Checks if the maximum number of connection switches has been reached or exceeded
+     * based on reason param.
+     *
+     * @param reason the reason for checking if we have reached max connection switched.
+     *  - if CONN_SPEED_SLOWEST_PART comparison will be done against
+     * mNumConnSwitchesSlowestPart
+     *  - if TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD comparison will be
+     * done against mNumConnSwitchesBelowSpeedThreshold
+     *  - if ON_RAIDED_ERROR we don't need to check any switch counter as any HTTP err
+     * is considered as permanent, which means that unused connection cannot be reused anymore. See
+     * unusedConnectionCanBeReused
+     *  - if ON_RECOVERABLE_RAIDED_ERROR comparison will be done against
+     * mNumConnSwitchesRecoverableError
+     *
+     * @return `true` if the maximum number of connection switches has been reached or
+     * exceeded, `false` otherwise.
+     */
+    bool maxUnusedConnSwitchesReached(const UnusedConn::ConnReplacementReason reason) const
+    {
+        switch (reason)
+        {
+            case UnusedConn::CONN_SPEED_SLOWEST_PART:
+                return mNumConnSwitchesSlowestPart >=
+                       DirectReadSlot::MAX_CONN_SWITCHES_SLOWEST_PART;
+            case UnusedConn::TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD:
+                return mNumConnSwitchesBelowSpeedThreshold >=
+                       DirectReadSlot::MAX_CONN_SWITCHES_BELOW_SPEED_THRESHOLD;
+            case UnusedConn::ON_RAIDED_ERROR:
+                return false;
+            case UnusedConn::ON_RECOVERABLE_RAIDED_ERROR:
+                return mNumConnSwitchesRecoverableError >=
+                       DirectReadSlot::MAX_CONN_SWITCHES_RECOVERABLE_ERR;
+        }
+        return false;
+    }
+
+    /**
+     * @brief increases counter for unused connection switches given a replacement reason
+     *
+     * @note: in case we need to replace unused conn by permanent failed raided part
+     * (UnusedConn::ON_RAIDED_ERROR), we don't need to increase any switch counter, as any HTTP err
+     * is considered as permanent, which means that unused connection cannot be reused anymore. See
+     * unusedConnectionCanBeReused
+     */
+    void increaseUnusedConnSwitches(const UnusedConn::ConnReplacementReason reason)
+    {
+        switch (reason)
+        {
+            case UnusedConn::CONN_SPEED_SLOWEST_PART:
+                ++mNumConnSwitchesSlowestPart;
+                return;
+            case UnusedConn::TRANSFER_OR_CONN_SPEED_UNDER_THRESHOLD:
+                ++mNumConnSwitchesBelowSpeedThreshold;
+                return;
+            case UnusedConn::ON_RAIDED_ERROR:
+                return;
+            case UnusedConn::ON_RECOVERABLE_RAIDED_ERROR:
+                ++mNumConnSwitchesRecoverableError;
+                return;
+        }
+    }
+
     /**
     *   @brief Adjust URL port in URL for streaming (8080).
     *
@@ -1073,7 +1230,7 @@ struct MEGA_API DirectRead
 struct MEGA_API DirectReadNode
 {
     handle h;
-    bool p;
+    bool isPublicHandle;
     string publicauth;
     string privateauth;
     string chatauth;
@@ -1113,8 +1270,42 @@ struct MEGA_API DirectReadNode
     // report failure to app and abort or retry all reads
     void retry(const Error &, dstime = 0);
 
-    DirectReadNode(MegaClient*, handle, bool, SymmCipher*, int64_t, const char*, const char*, const char*);
+    DirectReadNode(MegaClient* client,
+                   handle h,
+                   bool isPublicHandle,
+                   SymmCipher* symmCipher,
+                   int64_t ctriv,
+                   const char* privateAuth,
+                   const char* publicAuth,
+                   const char* chatAuth);
     ~DirectReadNode();
+
+    // Convenience.
+    template<typename Predicate>
+    static constexpr auto IsDirectReadPredicateV =
+        std::is_invocable_r_v<bool, Predicate, const DirectRead&>;
+
+    // Abort all reads satisfying a particular predicate.
+    template<typename Predicate>
+    auto abort(Predicate predicate) -> std::enable_if_t<IsDirectReadPredicateV<Predicate>>
+    {
+        // Iterate over all active reads.
+        for (auto i = reads.begin(); i != reads.end();)
+        {
+            // Get a pointer to the current read.
+            auto* read = *i++;
+
+            // Read doesn't satisfy our predicate.
+            if (!std::invoke(predicate, *read))
+                continue;
+
+            // Abort the read.
+            read->onFailure(API_EINCOMPLETE, retries, 0);
+
+            // And remove it from the list of active reads.
+            delete read;
+        }
+    }
 };
 
 } // namespace mega
