@@ -86,7 +86,7 @@ bool File::serialize(string *d) const
     d->append((char*)&ll, sizeof(ll));
     d->append(name.data(), ll);
 
-    auto tmpstr = getLocalname().platformEncoded();
+    auto tmpstr = getLocalname().serialize();
     ll = (unsigned short)tmpstr.size();
     d->append((char*)&ll, sizeof(ll));
     d->append(tmpstr.data(), ll);
@@ -123,7 +123,10 @@ bool File::serialize(string *d) const
 
     d->append((char*)&mCollisionResolution, 1);
 
-    d->append("\0\0\0\0\0\0\0", 8);
+    bool pathSerializedAsLocalPath = true;
+    d->append(reinterpret_cast<const char*>(&pathSerializedAsLocalPath), sizeof(bool));
+
+    d->append("\0\0\0\0\0\0", 7);
 
     if (hasChatAuth)
     {
@@ -230,7 +233,6 @@ File *File::unserialize(string *d)
     fp.reset();
 
     file->name.assign(name, namelen);
-    file->setLocalname(LocalPath::fromPlatformEncodedAbsolute(std::string(localname, localnamelen)));
     file->targetuser.assign(targetuser, targetuserlen);
     file->privauth.assign(privauth, privauthlen);
     file->pubauth.assign(pubauth, pubauthlen);
@@ -266,13 +268,33 @@ File *File::unserialize(string *d)
     }
     file->setCollisionResolution(static_cast<CollisionResolution>(collisionResolutionUint8));
 
-    if (memcmp(ptr, "\0\0\0\0\0\0\0", 8))
+    bool pathSerializedAsLocalPath = MemAccess::get<bool>(ptr);
+    ptr += sizeof(bool);
+
+    if (pathSerializedAsLocalPath)
+    {
+        auto localPath = LocalPath::unserialize(std::string(localname, localnamelen));
+        if (!localPath.has_value())
+        {
+            LOG_err << "File unserialization failed - invalid LocalPath";
+            delete file;
+            return NULL;
+        }
+        file->setLocalname(localPath.value());
+    }
+    else
+    {
+        file->setLocalname(
+            LocalPath::fromPlatformEncodedAbsolute(std::string(localname, localnamelen)));
+    }
+
+    if (memcmp(ptr, "\0\0\0\0\0\0", 7))
     {
         LOG_err << "File unserialization failed - invalid version";
         delete file;
         return NULL;
     }
-    ptr += 8;
+    ptr += 7;
 
     if (hasChatAuth)
     {
@@ -348,12 +370,22 @@ void File::completed(Transfer* t, putsource_t source)
 
     if (t->type == PUT)
     {
-        sendPutnodesOfUpload(t->client, t->uploadhandle, *t->ultoken, t->filekey, source, NodeHandle(), nullptr, nullptr, false);
+        sendPutnodesOfUpload(t->client,
+                             t->uploadhandle,
+                             "",
+                             *t->ultoken,
+                             t->filekey,
+                             source,
+                             NodeHandle(),
+                             nullptr,
+                             nullptr,
+                             false);
     }
 }
 
 void File::sendPutnodesOfUpload(MegaClient* client,
                                 UploadHandle fileAttrMatchHandle,
+                                std::string&& fileAttr,
                                 const UploadToken& ultoken,
                                 const FileNodeKey& newFileKey,
                                 putsource_t source,
@@ -370,7 +402,9 @@ void File::sendPutnodesOfUpload(MegaClient* client,
     newnode->canChangeVault = canChangeVault;
 
     // upload handle required to retrieve/include pending file attributes
+    // or file attribute value if it is not empty
     newnode->uploadhandle = fileAttrMatchHandle;
+    newnode->fileattributes = std::move(fileAttr);
 
     // reference to uploaded file
     newnode->uploadtoken = ultoken;
@@ -430,17 +464,17 @@ void File::sendPutnodesOfUpload(MegaClient* client,
             }
         }
 
-        client->reqs.add(new CommandPutNodes(client,
-                                             th,
-                                             NULL,
-                                             mVersioningOption,
-                                             std::move(newnodes),
-                                             tag,
-                                             source,
-                                             nullptr,
-                                             std::move(completion),
-                                             canChangeVault,
-                                             {})); // customerIpPort
+        client->queueCommand(new CommandPutNodes(client,
+                                                 th,
+                                                 NULL,
+                                                 mVersioningOption,
+                                                 std::move(newnodes),
+                                                 tag,
+                                                 source,
+                                                 nullptr,
+                                                 std::move(completion),
+                                                 canChangeVault,
+                                                 {})); // customerIpPort
     }
 }
 
@@ -495,17 +529,17 @@ void File::sendPutnodesToCloneNode(MegaClient* client,
         NodeHandle th = h;
         assert(syncxfer);
         newnode->ovhandle = ovHandle;
-        client->reqs.add(new CommandPutNodes(client,
-                                             th,
-                                             NULL,
-                                             mVersioningOption,
-                                             std::move(newnodes),
-                                             tag,
-                                             source,
-                                             nullptr,
-                                             std::move(completion),
-                                             canChangeVault,
-                                             {})); // customerIpPort
+        client->queueCommand(new CommandPutNodes(client,
+                                                 th,
+                                                 NULL,
+                                                 UseLocalVersioningFlag,
+                                                 std::move(newnodes),
+                                                 tag,
+                                                 source,
+                                                 nullptr,
+                                                 std::move(completion),
+                                                 canChangeVault,
+                                                 {})); // customerIpPort
     }
 }
 
@@ -515,8 +549,7 @@ void File::terminated(error)
 
 }
 
-// do not retry crypto errors or administrative takedowns; retry other types of
-// failuresup to 16 times, except I/O errors (6 times)
+// do not retry crypto errors or administrative takedowns
 bool File::failed(error e, MegaClient*)
 {
     if (e == API_EKEY)
@@ -524,15 +557,17 @@ bool File::failed(error e, MegaClient*)
         return false; // mac error; do not retry
     }
 
-    return  // Non fatal errors, up to 16 retries
-            ((e != API_EBLOCKED && e != API_ENOENT && e != API_EINTERNAL && e != API_EACCESS && e != API_ETOOMANY && transfer->failcount < 16)
-            // I/O errors up to 6 retries
-            && !((e == API_EREAD || e == API_EWRITE) && transfer->failcount > 6))
-            // Retry sync transfers up to 8 times for erros that doesn't have a specific management
-            // to prevent immediate retries triggered by the sync engine
-            || (syncxfer && e != API_EBLOCKED && e != API_EKEY && transfer->failcount <= 8)
-            // Infinite retries for storage overquota errors
-            || e == API_EOVERQUOTA || e == API_EGOINGOVERQUOTA;
+    return // Non fatal errors, up to FILE_MAX_RETRIES retries
+        ((e != API_EBLOCKED && e != API_ENOENT && e != API_EINTERNAL && e != API_EACCESS &&
+          e != API_ETOOMANY && transfer->failcount < FILE_MAX_RETRIES)
+         // I/O errors up to FILE_IO_MAX_RETRIES retries
+         && !((e == API_EREAD || e == API_EWRITE) && transfer->failcount > FILE_IO_MAX_RETRIES))
+        // Retry sync transfers up to FILE_SYNC_MAX_RETRIES times for erros that doesn't have a
+        // specific management to prevent immediate retries triggered by the sync engine
+        || (syncxfer && e != API_EBLOCKED && e != API_EKEY &&
+            transfer->failcount <= FILE_SYNC_MAX_RETRIES)
+        // Infinite retries for storage overquota errors
+        || e == API_EOVERQUOTA || e == API_EGOINGOVERQUOTA;
 }
 
 void File::displayname(string* dname)
@@ -605,6 +640,8 @@ void SyncUpload_inClient::completed(Transfer* t, putsource_t source)
     uploadHandle = t->uploadhandle;
     uploadToken = *t->ultoken;
     fileNodeKey = t->filekey;
+    if (auto c = t->client)
+        c->pendingattrstring(uploadHandle, &fileAttr);
 
     SyncTransfer_inClient::completed(t, source);
 }
@@ -629,6 +666,7 @@ void SyncUpload_inClient::sendPutnodesOfUpload(MegaClient* client, NodeHandle ov
     File::sendPutnodesOfUpload(
         client,
         uploadHandle,
+        std::move(fileAttr),
         uploadToken,
         fileNodeKey,
         PUTNODES_SYNC,
@@ -768,6 +806,10 @@ SyncUpload_inClient::SyncUpload_inClient(NodeHandle targetFolder, const LocalPat
 
     sourceFsid = fsid;
     sourceLocalname = localname;
+
+    LOG_debug << "[SyncUpload_inClient()] Name: '" << getLocalname() << "'. Source local name: '"
+              << sourceLocalname.toPath(false) << "'. Source fsid: " << fsid
+              << ". Fingerprint: " << fingerprintDebugString();
 }
 
 SyncUpload_inClient::~SyncUpload_inClient()
@@ -787,16 +829,14 @@ SyncUpload_inClient::~SyncUpload_inClient()
         syncThreadSafeState->transferFailed(PUT, size);
     }
 
-    if (!uploadHandle.isUndef())
-    {
-        // Remove file attributes if they weren't removed upon ~Transfer destructor
-        syncThreadSafeState->client()->fileAttributesUploading.erase(uploadHandle);
-    }
-
     if (putnodesStarted)
     {
         syncThreadSafeState->removeExpectedUpload(h, name);
     }
+
+    LOG_debug << "[~SyncUpload_inClient()] Name: '" << getLocalname() << "'. Source local name: '"
+              << sourceLocalname.toPath(false) << "'. Source fsid: " << sourceFsid
+              << ". Fingerprint: " << fingerprintDebugString();
 }
 
 void SyncUpload_inClient::prepare(FileSystemAccess&)
@@ -827,6 +867,11 @@ void SyncUpload_inClient::updateFingerprint(const FileFingerprint& newFingerprin
         syncThreadSafeState->transferBegin(PUT, newFingerprint.size);
     }
 
+    LOG_debug << "[SyncUpload_inClient::updateFingerprint] Name: '" << getLocalname()
+              << "'. Source fsid: " << sourceFsid
+              << ". Prev Fingerprint: " << fingerprintDebugString()
+              << ". New Fingerprint: " << newFingerprint.fingerprintDebugString();
+
     FileFingerprint::operator=(newFingerprint);
 }
 
@@ -844,6 +889,10 @@ SyncDownload_inClient::SyncDownload_inClient(CloudNode& n, const LocalPath& cloc
 
     syncThreadSafeState = std::move(stss);
     syncThreadSafeState->transferBegin(GET, size);
+
+    LOG_debug << "[SyncDownload_inClient()] Name: '" << getLocalname() << "'. Handle: " << h
+              << ". Cloud Fingerprint: " << fingerprintDebugString()
+              << ". Local Fingerprint (overwrite): " << overwriteFF.fingerprintDebugString();
 }
 
 SyncDownload_inClient::~SyncDownload_inClient()
@@ -865,6 +914,10 @@ SyncDownload_inClient::~SyncDownload_inClient()
     {
         syncThreadSafeState->transferFailed(GET, size);
     }
+
+    LOG_debug << "[~SyncDownload_inClient()] Name: '" << getLocalname() << "'. Handle: " << h
+              << ". Cloud Fingerprint: " << fingerprintDebugString()
+              << ". Local Fingerprint (overwrite): " << okToOverwriteFF.fingerprintDebugString();
 }
 
 void SyncDownload_inClient::prepare(FileSystemAccess&)
